// ################################################################ InitGraphics
// Initialisiert alle grafischen Elemente wie Bitmaps und Controls
procedure TMainForm.InitGraphics;
var i:integer;
begin
  if not GraphicsInitialized then begin
    Back:=TBitmap.Create;
    with Back do begin
      LoadFromFile(BackgroundFileName);
      BackSizeX:=Width;
      BackSizeY:=Height;
    end;
    Buf:=TBitmap.Create;
    Buf.LoadFromFile(TileSetFileName);
    with Buf.Canvas do begin
      with Pen do begin
        Style:=psSolid;
        Width:=7;
        Color:=$FF;
      end;
      Brush.Style:=bsClear;
      with Font do begin
        if Screen.Fonts.IndexOf(PreferredFont)>=0
          then Name:=PreferredFont
          else Name:='Tahoma';
        Size:=BigFontSize;
        Style:=[fsBold];
      end;
    end;
    Sprites:=TBitmap.Create;
    with Sprites do begin
      LoadFromFile(TileSetFileName);
      TransparentColor:=0;
      TransparentMode:=tmFixed;
    end;
    for i:=0 to 42 do
      SpriteRect[i]:=Rect(i*TileSizeX,0,(i+1)*TileSizeX,TileSizeY);
  end else
    Field.Free;

  BufSizeX:=FieldSizeX*TileSizeX+2*FieldBorder;
  BufSizeY:=FieldSizeY*TileSizeY+2*FieldBorder;
  ClientWidth:=BufSizeX+2;       ClientHeight:=BufSizeY+29;
  FieldBevel.Width:=BufSizeX+2;  FieldBevel.Height:=BufSizeY+2;
  Buf.Width:=BufSizeX;           Buf.Height:=BufSizeY;
  Field:=TImage.Create(Self);
  with Field do begin
    Parent:=Self;
    Left:=1; Top:=28;
    Width:=BufSizeX;
    Height:=BufSizeY;
    OnClick:=FieldClick;
    OnMouseDown:=FieldMouseDown;
    OnMouseUp:=FieldMouseUp;
  end;
  GraphicsInitialized:=true;
end;

// ###################################################################### Redraw
// zeichnet das Spielfeld neu
procedure TMainForm.Redraw;
var X,Y,PX,PY,Sprite:integer;
begin
  Y:=0;while Y<BufSizeY do begin
    X:=0;while X<BufSizeX do begin
      Buf.Canvas.Draw(X,Y,Back);
    inc(X,BackSizeX);end;
  inc(Y,BackSizeY);end;
  PY:=FieldBorder;
  for Y:=1 to FieldSizeY do begin
    PX:=FieldBorder;
    for X:=1 to FieldSizeX do begin
      Sprite:=FieldData[X,Y];
      if Sprite>0 then begin
        Buf.Canvas.CopyRect(Rect(PX,PY,PX+TileSizeX,PY+TileSizeY),
          Sprites.Canvas,SpriteRect[Sprite]);
        if (X=SelTileX) AND (Y=SelTileY) then Enlightenment(X,Y);
      end;
      inc(PX,TileSizeX);
    end;
    inc(PY,TileSizeY);
  end;
  Flip;
end;

// ######################################################################## Flip
// zeigt die Spielfeld-Bitmap an
procedure TMainForm.Flip;
begin
  Field.Canvas.Draw(0,0,Buf);
  if ForceRepaint then Field.Repaint;
end;

// ############################################################### Enlightenment
// hebt ein Tile hervor
procedure TMainForm.Enlightenment(TileX,TileY:integer);
var X,Y,PX,PY:integer; BMPpos:^byte;
begin
  PX:=(TileX-1)*TileSizeX+FieldBorder;
  PY:=(TileY-1)*TileSizeY+FieldBorder;
  for Y:=(PY+TileSizeY-1) downto PY do begin
    BMPpos:=@Buf.ScanLine[Y]^;
    inc(BMPpos,PX);
    for X:=TileSizeX downto 1 do begin
      if BMPpos^=EnlightenmentSrcColor then BMPpos^:=EnlightenmentDestColor;
      inc(BMPpos);
    end;
  end;
end;

// ################################################################# DrawBigText
// zeichnet einen groen, fetten Text in die Spielfeldmitte
procedure TMainForm.DrawBigText(const txt:string;col:TColor);
var X,Y:integer; S:TSize;
begin
  with Buf.Canvas do begin
    S:=TextExtent(txt);
    X:=(BufSizeX-S.CX-ShadowOffset SHR 1) DIV 2;
    Y:=(BufSizeY-S.CY-(ShadowOffset+BigFontSize) SHR 1) DIV 2;
    Font.Color:=0;
    TextOut(X+ShadowOffset,Y+ShadowOffset,txt);
    Font.Color:=col;
    TextOut(X,Y,txt);
  end;
end;

// ########################################################### Field.OnMouseDown
// Eine Maustaste wird ber dem Spielfeld gedrckt -> Tile hervorheben
procedure TMainForm.FieldMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var TX,TY:integer;
begin
  if Paused OR (MovesLeft=0) then exit;
  TX:=(X-FieldBorder+TileSizeX) DIV TileSizeX;
  TY:=(Y-FieldBorder+TileSizeY) DIV TileSizeY;
  if (TX<1) OR (TX>FieldSizeX) OR (TY<1) OR (TY>FieldSizeY) then exit;
  Enlightenment(TX,TY);Flip;
  ClickTileX:=TX;ClickTileY:=TY;
end;

// ############################################################# Field.OnMouseUp
// der ultimative Spielfeld-Maushandler
procedure TMainForm.FieldMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var TX,TY,i,Sprite:integer;
begin
  if Paused OR (MovesLeft=0) then exit;
  TX:=(X-FieldBorder+TileSizeX) DIV TileSizeX;
  TY:=(Y-FieldBorder+TileSizeY) DIV TileSizeY;

  // Falle 1: Rausschmi bei ...
  if (TX<1) OR (TX>FieldSizeX) OR (TY<1) OR (TY>FieldSizeY) OR  // Klick auerhalb des Spielfelds
     (TX<>ClickTileX) OR (TY<>ClickTileY) OR  // Verrutschen mit der Maus seit Festhalten der Taste
     (FieldData[TX,TY]=0) then begin  // Klick auf ein leeres Tile
    SelTileX:=0; SelTileY:=0;
    Redraw; exit;
  end;

  // Falle 2: Rechte Maustaste dient zum Hervorheben gleicher Tiles
  if (Button=mbRight) then begin
    SelTileX:=0; SelTileY:=0;
    Redraw;
    Sprite:=FieldData[TX,TY];
    for i:=1 to TileCount[Sprite] do
      with TilePos[Sprite,i] do
        if FieldData[X,Y]>0 then
          Enlightenment(X,Y);
    Flip; exit;
  end;

  // Falle 3: noch kein Tile markiert -> Tile markieren
  if (SelTileX=0) AND (SelTileY=0) then begin
    SelTileX:=TX; SelTileY:=TY; exit;
  end;

  // Falle 4: bereits markiertes Tile angeklick -> Markierung aufheben
  if (TX=SelTileX) AND (TY=SelTileY) then begin
    SelTileX:=0; SelTileY:=0;
    Redraw;exit;
  end;

  // Falle 5: zwei verschiedene Tiles angeklickt, aber Zug ist nicht mglich
  //          -> Markierung auf zuletzt angeklicktes Tile setzen
  if not MoveCheck(SelTileX,SelTileY,TX,TY,true) then begin
    SelTileX:=TX; SelTileY:=TY;
    Redraw; exit;
  end;

  // Alle Fallen berstanden? Dann Zug ausfhren
  DoMove(SelTileX,SelTileY,TX,TY);
end;

// ###################################################################### DoMove
// Fhrt einen Zug aus, d.h. mit History-Eintrag, Anzeige-Timer usw.
procedure TMainForm.DoMove(X1,Y1,X2,Y2:integer);
begin
  HintAvailable:=false;
  dec(TilesLeft,2);
  inc(HistPos);HistMax:=HistPos;
  with History[HistMax,1] do begin X:=X1;Y:=Y1;Sprite:=FieldData[X1,Y1];end;
  with History[HistMax,2] do begin X:=X2;Y:=Y2;Sprite:=FieldData[X2,Y2];end;
  FieldData[X1,Y1]:=0;
  FieldData[X2,Y2]:=0;
  SelTileX:=0; SelTileY:=0;
  ShowMoveTimer.Enabled:=true;
  EvaluateGame;
end;

// ################################################################ EvaluateGame
// berprft die aktuelle Spielsituation auf Anzahl mglicher Zge
// sowie Lsbarkeit und trgt die Ergebnisse ins Info-Panel ein;
// auerdem wird hier die Gewinner-/Verlierer-Behandlung implementiert
procedure TMainForm.EvaluateGame;
begin
  if Solvable then
    InfoSolvable.Caption:='a'
  else if (TilesMax>84) then
    InfoSolvable.Caption:='s'
  else
    InfoSolvable.Caption:='r';
  MovesLeft:=GetMoves;
  InfoMovesLeft.Caption:=IntToStr(MovesLeft);
  if MovesLeft=0 then begin
    ShowMoveTimer.Enabled:=false;
    Redraw;
    if TilesLeft<=0 then begin
      Clock.Enabled:=false;
      DrawBigText('GEWONNEN!',clLime);Flip;
      if not InputQuery('Sie haben gewonnen!',
                        'Bitte geben Sie Ihren Namen fr die Bestenliste ein:',
                        PlayerName) then PlayerName:='';
      if (PlayerName='') then PlayerName:='Mr. Shy';
      ScoreForm.AddScore(PlayerName,FieldSizeX,FieldSizeY,TimeSec);
      ScoreForm.ShowModal;
    end else begin
      DrawBigText('VERLOREN!',clRed);Flip;
    end;
    exit;
  end;
end;

